#include <assert.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include "graphics.h"
#include "map.h"
#include "misc.h"
#include "pud.h"

spriteset spritesets[256];
spriteset fontset;
tileset tilesets[256];

BITMAP *vscreen;
BITMAP *background;

RGB palette[256];

void d_sprite(BITMAP *dest, int set, int num, int x, int y)
{ assert(spritesets[set].nsprites>num);
  draw_sprite(dest, spritesets[set].sprites[num], x, y);
}

void printfxy(int x, int y, char *fmt, ...)
{ va_list args;
  static char *lookup="abcdefghijklmnopqrstuvwxyz0123456789:(),-", str[256];
  char *s=str;

  va_start(args, fmt);
  vsprintf(str, fmt, args);
  va_end(args);

  for(; *s; s++, x+=16)
    if(strchr(lookup, *s))
      draw_sprite(vscreen, fontset.sprites[strchr(lookup, *s)-lookup], x, y);
}

void flush()
{ blit(vscreen, screen, 0, 0, 0, 0, 640, 480);
}

int translucent(int t)
{ return t==WATER || t==BLINK_BRICK;
}

inline void getsurrounding(int surrounding[8], int layer, int x, int y)
{ if(x-1>=0) surrounding[0]=map[layer][x-1][y];
  if(x-1>=0 && y-1>=0) surrounding[1]=map[layer][x-1][y-1];
  if(y-1>=0) surrounding[2]=map[layer][x][y-1];
  if(x+1<=255 && y-1>=0) surrounding[3]=map[layer][x+1][y-1];
  if(x+1<=255) surrounding[4]=map[layer][x+1][y];
  if(x+1<=255 && y+1<=255) surrounding[5]=map[layer][x+1][y+1];
  if(y+1<=255) surrounding[6]=map[layer][x][y+1];
  if(x-1>=0 && y+1<=255) surrounding[7]=map[layer][x-1][y+1];
}


inline int findsubtile(int layer, int x, int y)
{ int surrounding[8]={ 0, 0, 0, 0, 0, 0, 0, 0 };
  int tile=map[layer][x][y];

  switch(tile&255)
  { case STRIPED:
    case GRASS:
    case REDBLOCK:
    case GREENBLOCK:
    case BLUEBLOCK:
    case WHITEBLOCK:
    case PURPLEBLOCK:
      getsurrounding(surrounding, layer, x, y);
      return (surrounding[4]==tile)+
             (surrounding[6]==tile)*2+
             (surrounding[0]==tile)*4+
             (surrounding[2]==tile)*8;
    case STRIPEDBG:
      if(map[0][x][y]==STRIPED) return 8;
      getsurrounding(surrounding, 0, x, y);
      return (surrounding[2]==STRIPED)+
             (surrounding[3]==STRIPED)*2+
             (surrounding[4]==STRIPED)*4;
  }

  return 0;
}

void drawscreen(int x, int y, int drawbackground,
                              int drawlayer0, int drawlayer1)
{ int i, j, k;

  if(drawbackground) blit(background, vscreen, x/32, y/32, 0, 0, 640, 480);
  else clear(vscreen);

  for(i=x/32; i<x/32+21; i++)
    for(j=y/32; j<y/32+16; j++)
      for(k=drawlayer1; k>=1-drawlayer0; k--)
      { int tileset=map[k][i][j];
        int subtile;

        if(map[k][i][j]>>8) subtile=map[k][i][j]>>8;
        else subtile=findsubtile(k, i, j);

        tileset&=255;

        if(tileset && i*32+32>x && i*32-32<x+640
                   && j*32+32>y && j*32-32<y+480)
        { if(translucent(tileset))
            draw_trans_sprite(vscreen, tilesets[tileset].tiles[subtile],
                              i*32-x, j*32-y);
          else
            draw_sprite(vscreen, tilesets[tileset].tiles[subtile],
                        i*32-x, j*32-y);
        }
      }
}

BITMAP *grab_bitmap(BITMAP *src, int x, int y, int width, int height)
{ BITMAP *b=create_bitmap(width, height);
  int i, j;

  for(i=0; i<height; i++)
    for(j=0; j<width; j++)
      b->line[i][j]=src->line[i+y][j+x];

  return b;
}

tileset loadtileset(char *fn)
{ tileset s;
  int i, j;
  unsigned char *data=(unsigned char*)loadlump(fn);
  unsigned char *datap=data;

  data+=4;
  s.ntiles=decodelong(data);
  data+=4;
  assert(s.tiles=(BITMAP **)calloc(s.ntiles, sizeof(BITMAP *)));
  assert(s.type=(int *)calloc(s.ntiles, sizeof(int)));

  for(i=0; i<s.ntiles; i++)
  { s.tiles[i]=create_bitmap(32, 32);
    for(j=0; j<32; j++)
    { memcpy(s.tiles[i]->line[j], data, 32);
      data+=32;
    }
  }

  free(datap);
  return s;
}

void loadtiles()
{ int i;

  tilesets[1]=loadtileset("tiles.brick");
  tilesets[2]=loadtileset("tiles.water");
  tilesets[3]=loadtileset("tiles.grass");
  tilesets[4]=loadtileset("tiles.oneway");
  tilesets[5]=loadtileset("tiles.blinkbricks");
  tilesets[6]=loadtileset("tiles.flora");
  tilesets[7]=loadtileset("tiles.striped");
  tilesets[8]=loadtileset("tiles.stripedbg");
  tilesets[9]=loadtileset("tiles.redblock");
  tilesets[10]=loadtileset("tiles.greenblock");
  tilesets[11]=loadtileset("tiles.blueblock");
  tilesets[12]=loadtileset("tiles.whiteblock");
  tilesets[13]=loadtileset("tiles.purpleblock");
  for(i=0; i<tilesets[WATER].ntiles; i++) tilesets[2].type[i]=UNBLOCKING;
  for(i=0; i<6; i++)
  { tilesets[ONE_WAY].type[i]=ONEWAYLEFT;
    tilesets[ONE_WAY].type[i+6]=ONEWAYRIGHT;
  }
  tilesets[BLINK_BRICK].type[0]=UNBLOCKING;
  for(i=0; i<tilesets[6].ntiles; i++) tilesets[6].type[i]=UNBLOCKING;
}

spriteset loadspriteset(char *fn)
{ spriteset s;
  int i, j;
  unsigned char *data=(unsigned char*)loadlump(fn);
  unsigned char *datap=data;

  data+=4;
  s.nsprites=decodelong(data);
  data+=4;
  assert(s.sprites=(BITMAP **)calloc(s.nsprites, sizeof(BITMAP **)));

  for(i=0; i<s.nsprites; i++)
  { short x, y;

    x=decodeword(data);
    data+=2;
    y=decodeword(data);
    data+=2;
    s.sprites[i]=create_bitmap(x, y);

    for(j=0; j<y; j++)
    { memcpy(s.sprites[i]->line[j], data, x);
      data+=x;
    }
  }

  free(datap);
  return s;
}

void loadsprites()
{ spritesets[0]=loadspriteset("sprites.cursor");
  spritesets[1]=loadspriteset("sprites.swarmie");
  spritesets[2]=loadspriteset("sprites.blob");
  spritesets[3]=loadspriteset("sprites.spikes");
  spritesets[4]=loadspriteset("sprites.keys");
  spritesets[5]=loadspriteset("sprites.locks");
  spritesets[6]=loadspriteset("sprites.coin");
  spritesets[7]=loadspriteset("sprites.breaker");
  spritesets[8]=loadspriteset("sprites.spring");
  spritesets[9]=loadspriteset("sprites.portcullis");
  spritesets[10]=loadspriteset("sprites.float");
  spritesets[11]=loadspriteset("sprites.bubble");

  fontset=loadspriteset("sprites.font");
}

void loadtranstable()
{ COLOR_MAP *light_table;
  light_table=(COLOR_MAP *)loadlump("colormap");
  color_map=light_table;
}

void loadbackground()
{ background=load_bitmap("bg.pcx", palette);
}
